"""
Copyright (C) 2022 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from datetime import datetime
from tabnanny import verbose

from attr import has
from domonic.ext.html5lib_ import getTreeBuilder
from domonic.html import a, b, button, div, pre, script, style, table, td, th, tr
from domonic.utils import Utils
import html5lib
import inspect
import io
import pandas
import pkg_resources
import pprint
import re
import sys


import kaskada.fenl.v1alpha as kaskada_fenl_pb2

kaskada_fenl_primitive_types = {
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_BOOL: "bool",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_MICROSECOND: "duration_us",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_MILLISECOND: "duration_ms",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_NANOSECOND: "duration_ns",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_SECOND: "duration_s",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_F16: "f16",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_F32: "f32",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_F64: "f64",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I16: "i16",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I32: "i32",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I64: "i64",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I8: "i8",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_INTERVAL_DAY_TIME: "interval_days",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_INTERVAL_YEAR_MONTH: "interval_months",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_NULL: "null",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_STRING: "string",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_MICROSECOND: "timestamp_us",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_MILLISECOND: "timestamp_ms",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_NANOSECOND: "timestamp_ns",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_SECOND: "timestamp_s",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U16: "u16",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U32: "u32",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U64: "u64",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U8: "u8",
}

def convert_fenl_schema_field(field):
    """
    Converts a fenl field to a string representation.

    Args:
        field (kaskada_fenl_pb2.schema_pb2.Schema_Field) : the input field

    Returns:
        str: The string representation of the fenl Field
    """
    return field.name, convert_fenl_datatype(field.data_type)

def convert_fenl_datatype(data_type):
    """
    Converts a fenl dataType to a string representation.

    Args:
        data_type (kaskada_fenl_pb2.schema_pb2.DataType) : the input dataType

    Returns:
        str: The string representation of the fenl DataType
    """
    type_string = None
    if data_type.HasField("primitive"):
        type_string = kaskada_fenl_primitive_types.get(data_type.primitive, "<unknown>")
    elif data_type.HasField("struct"):
        type_string = "<struct>"
    elif data_type.HasField("window"):
        type_string = "<window>"
    return type_string

def get_classname(obj):
    return f"{obj.__module__}.{obj.__name__}"

def get_schema_dataframe(obj):
    """
    Gets a dataframe from a schema object if it exists

    Args:
        obj (kaskada.fenl.v1alpha.schema_pb2.Schema | Any): Either a schema object or an object that has a schema object at the root level

    Returns:
        pandas.dataframe | None: Either a dataframe describing the schemea or None if no schema exists in the passed object.
    """
    if hasattr(obj, "schema"):
        obj = obj.schema

    if (not hasattr(obj, "fields")) or len(obj.fields) == 0:
        return None
    fields = {}
    for idx, field in enumerate(obj.fields):
        name, typeString = convert_fenl_schema_field(field)
        fields[idx] = [name, typeString]
    return pandas.DataFrame.from_dict(data=fields, orient="index", columns=["column_name", "column_type"])

def get_datetime(pb2_timestamp):
    """
    Converts a google.protobuf.timestamp_pb2 into a python datetime object.

    Args:
        pb2_timestamp (google.protobuf.timestamp_pb2): The protobuf timestamp.

    Returns:
        datetime: The python datetime.
    """
    return datetime.fromtimestamp(pb2_timestamp.seconds + pb2_timestamp.nanos/1e9)

def get_properties(obj):
    """
    Converts a protobuf response into a simple python dictionary.

    Args:
        obj (Any): A deserialized protobuf object.

    Returns:
        dict of {str : Any}: A nested set of keys and values with structure matching the protobuf input.
    """
    results = {}
    for propName in dir(obj):
        if re.search(r"^[a-z]", propName) is None:
            continue
        propVal = getattr(obj,propName)
        propType = type(propVal)
        prop_class = get_classname(propType)
        if propType.__name__ in ("method"):
            continue
        elif propType in (int, float, bool):
            results[propName] = propVal
        elif propType is str:
            if propVal != "":
                results[propName] = propVal
        elif prop_class == "google.protobuf.pyext._message.RepeatedCompositeContainer":
            if len(propVal) > 0:
                results[propName] = [get_properties(subObj) for subObj in propVal]
        elif prop_class == "google.protobuf.timestamp_pb2.Timestamp":
            results[propName] = get_datetime(propVal).isoformat()
        elif prop_class.startswith("pandas."):
            results[propName] = propVal
        else:
            results[propName] = get_properties(propVal)
    return results

def get_item(obj, keys, default = ""):
    """
    Safely recursively gets an item from a dict, otherwise returns the default value

    Args:
        obj (dict of {str : Any}): Dictionary of (nested) key item pairs
        keys (list of str): The list of keys use to navigate to the desired item in the dictionary.
        default (str, optional): The default value to return if the item isn't found. Defaults to "".

    Returns:
        string: Either the found value or the default.
    """
    while (len(keys) > 0):
        key = keys.pop(0)
        if key in obj:
            obj = obj[key]
        else:
            return default
    return obj

def appendChildIfNotNone(parent, child):
    if child is not None:
        parent.appendChild(child)

def appendChildren(parent, children):
    for child in children:
        appendChildIfNotNone(parent, child)

def html_obj_table_row(key, value):
    """
    Returns a html table row, where the key is bolded, and the value is output as code.
    """
    if value is None or value == "":
        return None
    return tr(td(b(key)),td(pre(value)))

def html_table_row(key, value):
    """
    Returns a html table row, where the key is bolded, and the value is output as is.
    """
    return tr(td(b(key)),td(value))

def get_request_details_table_row_if_exists(props):
    if "request_details" in props:
        request_details = props.pop("request_details")
        nested_table = table(html_obj_table_row("request_id", get_item(request_details, ["request_id"])), _class="kda_table")
        return html_table_row("request_details", nested_table)
    return None

def html_obj_id_row(index, obj, props):
    obj_type = type(obj).__name__.lower()
    obj_name_key = f"{obj_type}_name"
    obj_name = get_item(props, [obj_name_key])
    return tr(td(pre(index)),td(pre(obj_name)))

def get_shared_generic_response_props_and_schema_dataframe(obj):
    props = get_properties(obj)
    props.pop("file_count", "") # remove table `file_count` since it isn't populated yet
    schema_df = None
    # special handling of `result_type` for views
    if hasattr(obj, "result_type"):
        fenl_datatype = convert_fenl_datatype(obj.result_type)
        if fenl_datatype is None:
            props.pop("result_type","") #remove `result_type` from props
        else:
            props["result_type"] = Utils.escape(fenl_datatype)
            if obj.result_type.HasField("struct"):
                schema_df = get_schema_dataframe(obj.result_type.struct)
                if schema_df is not None:
                    props["result_type"] += " (see Schema tab)" 

    # remove schema from details view, put in schema viewer instead
    if hasattr(obj, "schema"):
        props.pop("schema", "")
        schema_df = get_schema_dataframe(obj)
        if schema_df is not None:
            props["schema"] = "(see Schema tab)" 

    return props, schema_df

def get_shared_generic_response_table_rows(obj, props):
    obj_type = type(obj).__name__.lower()

    # try to pop off obj_id and obj_name first
    obj_id_key = f"{obj_type}_id"
    obj_id = props.pop(obj_id_key, None)
    obj_name_key = f"{obj_type}_name"
    obj_name = props.pop(obj_name_key, None)

    # also remove create & update time, and slice info from props
    create_time = props.pop("create_time", None)
    update_time = props.pop("update_time", None)
    props.pop("slice", None)

    rows = [
        # first row is item name
        html_obj_table_row(obj_name_key, obj_name),
        # stop showing item ids since they aren't actionable by the user
        # html_obj_table_row(obj_id_key, obj_id),
    ]

    # add remainder of fields
    for key in props:
        rows.append(html_obj_table_row(key, props[key]))

    # special handling for slice configs
    if hasattr(obj, "slice"):
        rows.append(html_table_row("slice", slice_request_table(obj.slice)))
    
    # add create & update time at the end
    rows.append(html_obj_table_row("create_time", create_time))
    rows.append(html_obj_table_row("update_time", update_time))

    return rows

def generic_response_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    
    # geneneric responses should contain 2 properties at the root level
    # request_details and the actual object
    props = get_properties(obj)
    details = table(_class="kda_table")

    # pull off request_details
    request_details = get_request_details_table_row_if_exists(props)

    # pull of analysis
    errors = None
    can_execute = None
    analysis_props = props.pop("analysis", None)
    if analysis_props is not None:
        can_execute, nested_details, errors = analysis_and_errors_tables(obj.analysis)
        details.appendChild(html_table_row("analysis", nested_details))

    schema_df = None
    if can_execute is not False:
        # expect that only the actual object is left
        if len(props) == 1 and type(props[list(props)[0]]) not in (int, float, bool, str):
            obj_key = list(props)[0]
            actual_obj = getattr(obj, obj_key)
            
            actual_props, schema_df = get_shared_generic_response_props_and_schema_dataframe(actual_obj)
            if len(actual_props) > 0:
                nested_table = table(_class="kda_table")
                appendChildren(nested_table, get_shared_generic_response_table_rows(actual_obj, actual_props))
                details.appendChild(html_table_row(obj_key, nested_table))
        else:
            appendChildren(details, get_shared_generic_response_table_rows(obj, props))

    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None 
    appendChildIfNotNone(details, request_details)

    title_to_set_active = "Details" if errors is None else "Errors"
    return str(tab_panel([("Details", details), ("Schema", schema), ("Errors", errors), ("Raw", pre(pprint.pformat(obj)))],title_to_set_active=title_to_set_active))

def generic_object_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    
    props, schema_df = get_shared_generic_response_props_and_schema_dataframe(obj)
    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None

    details = table(_class="kda_table")
    appendChildren(details, get_shared_generic_response_table_rows(obj, props))

    return str(tab_panel([("Details", details), ("Schema", schema), ("Raw", pre(pprint.pformat(obj)))]))

def schema_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    schema_df = get_schema_dataframe(obj)
    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None

    return str(tab_panel([("Schema", schema), ("Raw", pre(pprint.pformat(obj)))]))

def data_type_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    title_to_set_active = ""
    result_type = Utils.escape(convert_fenl_datatype(obj))
    schema_df = None
    if obj.HasField("struct"):
        schema_df = get_schema_dataframe(obj.struct)
        if schema_df is not None:
            result_type += " (see Schema tab)"
            title_to_set_active = "Schema"
    
    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None

    return str(tab_panel([("Result Type", pre(result_type)), ("Schema", schema), ("Raw", pre(pprint.pformat(obj)))], title_to_set_active))

def response_delete_html_formatter(del_resp):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    
    request_details = get_request_details_table_row_if_exists(get_properties(del_resp))

    del_resp_type = type(del_resp).__name__
    del_obj = del_resp_type.replace("Response", "")
    
    details = table(_class="kda_table")
    details.appendChild(html_obj_table_row(del_obj, "success"))
    details.appendChild(request_details)

    return str(tab_panel([("Details", details), ("Raw", pre(pprint.pformat(del_resp)))]))

def response_list_html_formatter(list_resp):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    props = get_properties(list_resp)

    request_details = get_request_details_table_row_if_exists(props)

    list_resp_type = type(list_resp).__name__
    list_key = list_resp_type.replace("Response", "").replace("List", "").lower()

    details = table(_class="kda_table")
    
    # after pulling the request_id, only the item list should remain
    if len(props) == 1:
        list_value = getattr(list_resp, list_key)
        list_props = props[list_key]
        list_type = type(list_value)
        nested_table = table(_class="kda_table")
        details.appendChild(html_table_row(list_key, nested_table))
        nested_table.appendChild(tr(th("index"),th("name")))
        
        if get_classname(list_type) == "google.protobuf.pyext._message.RepeatedCompositeContainer":
            for i in range(len(list_value)):
                appendChildIfNotNone(nested_table, html_obj_id_row(str(i), list_value[i], list_props[i]))
    else: # when list is empty
        details.appendChild(html_table_row(list_key, Utils.escape("<empty>")))

    
    appendChildIfNotNone(details, request_details)

    return str(tab_panel([("Details", details), ("Raw", pre(pprint.pformat(list_resp)))]))

def proto_list_html_formatter(list_resp):
    item_count = len(list_resp)
    if item_count == 0:
        return str(pre(pprint.pformat(list_resp)))

    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    details = table(_class="kda_table")

    # peek at the type of items in the list
    item_class = get_classname(type(list_resp[0]))

    if item_class == "kaskada.errdetails.v1alpha.fenl_diagnostics_pb2.FenlDiagnostic":
        for i in range(item_count):
            fenl_diag = list_resp[i]
            details.appendChild(html_obj_table_row(i, fenl_diag.formatted))
    else:
        details.appendChild(tr(th("index"),th("name")))
        for i in range(item_count):
            appendChildIfNotNone(details, html_obj_id_row(i, list_resp[i], get_properties(list_resp[i])))

    return str(tab_panel([("Details", details), ("Raw", pre(pprint.pformat(list_resp)))]))

def slice_request_table(obj):
    details = table(_class="kda_table")
    if obj.HasField("percent"):
        details.appendChild(html_obj_table_row("percent", obj.percent.percent))
    else:
        details.appendChild(html_obj_table_row("None", "(full dataset used for query)"))
    return details

def slice_request_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    return str(tab_panel([("Details", slice_request_table(obj)), ("Raw", pre(pprint.pformat(obj)))]))

def get_shared_query_response_content(resp_obj):
    props = get_properties(resp_obj)
    parquet_path = get_item(props, ["parquet", "path"])
    redis_bulk_path = get_item(props, ["redis_bulk", "path"])

    resultsExist = False
    details = table(_class="kda_table")
    if parquet_path != "":
        nested_table = table(html_table_row("path",a(parquet_path, _href=parquet_path)), _class="kda_table")
        details.appendChild(html_table_row("parquet", nested_table))
        resultsExist = True
    elif redis_bulk_path != "":
        nested_table = table(html_table_row("path",a(redis_bulk_path, _href=redis_bulk_path)), _class="kda_table")
        details.appendChild(html_table_row("redis_bulk", nested_table))
        resultsExist = True
    else:
        nested_table = table(html_obj_table_row("can_execute", get_item(props, ["analysis", "can_execute"])), _class="kda_table")
        details.appendChild(html_table_row('analysis', nested_table))

    if resultsExist:
        details.appendChild(html_table_row("slice", slice_request_table(resp_obj.slice)))

    schema = None
    schema_df = get_schema_dataframe(resp_obj)
    if schema_df is not None:
        schema = convert_df_to_domonic(schema_df)
        details.appendChild(html_obj_table_row("schema", "(see Schema tab)"))

    rows = [
        html_obj_table_row("data_token_id", get_item(props, ["data_token_id"])),
        html_obj_table_row("next_resume_token", get_item(props, ["next_resume_token"])),
        get_request_details_table_row_if_exists(props),
    ]
    appendChildren(details, rows)

    return { "Details": details, "Schema": schema}

def query_response_html_formatter(resp_obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    
    content = get_shared_query_response_content(resp_obj)

    return str(tab_panel([("Details", content["Details"]), ("Schema", content["Schema"]), ("Raw", pre(pprint.pformat(resp_obj)))]))

def fenlmagic_query_result_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    content = get_shared_query_response_content(obj.query_response)
    details_table = content["Details"]
    details_table.appendChild(html_obj_table_row("query", obj.query))

    raw = {
        "query": obj.query,
        "query_response": obj.query_response,
    }

    dataframe = None
    if obj.dataframe is not None:
        buffer = io.StringIO()
        obj.dataframe.info(buf=buffer, verbose=False)
        raw["dataframe"] = buffer.getvalue().replace("<class 'pandas.core.frame.DataFrame'>\n","")
        dataframe = convert_df_to_domonic(obj.dataframe, 10, True)

    return str(tab_panel([("Dataframe", dataframe), ("Details", content["Details"]), ("Schema", content["Schema"]), ("Raw", pre(pprint.pformat(raw).replace("\\n","")))]))

def entity_filter_html_formatter(obj):
    props = get_properties(obj)
    details = table(_class="kda_table")
    filter_type = type(obj).__name__
    if filter_type == "EntityPercentFilter":
        details.appendChild(html_obj_table_row("percent", obj.percent))

    return str(tab_panel([("Details", details), ("Raw", pre(pprint.pformat(props).replace("\\n","")))]))

def analysis_and_errors_tables(obj):
    details = table(html_obj_table_row("can_execute", str(obj.can_execute)), _class="kda_table")
    errors = None
    if not obj.can_execute:
        details.appendChild(html_obj_table_row("fenl_diagnostics", "(see Errors tab)"))
        errors = fenl_diagnostics_table(obj.fenl_diagnostics)
    return obj.can_execute, details, errors

def analysis_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    can_execute, details, errors = analysis_and_errors_tables(obj)
    title_to_set_active = "Details" if errors is None else "Errors"
    return str(tab_panel([("Details", details), ("Errors", errors), ("Raw", pre(pprint.pformat(obj)))], title_to_set_active=title_to_set_active))

def fenl_diagnostics_table(obj):
    details = table(_class="kda_table")
    for i in range(len(obj.fenl_diagnostics)):
        fenl_diag = obj.fenl_diagnostics[i]
        details.appendChild(html_obj_table_row(i, fenl_diag.formatted))
    return details

def fenl_diagnostics_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    details = table(html_table_row("fenl_diagnostics", fenl_diagnostics_table(obj)))
    return str(tab_panel([("Details", details), ("Raw", pre(pprint.pformat(obj)))]))

def convert_df_to_domonic(df, max_rows = None, show_dimensions = False):
    parser = html5lib.HTMLParser(tree=getTreeBuilder())
    return parser.parse(df.to_html(max_rows=max_rows, notebook=True, show_dimensions=show_dimensions))
        
def tab_panel(content_list, title_to_set_active = ""):
    """
    Creates a tab-panel for the passed content list.

    Args:
        content_list (list of tuple of (str, Any)): A list of tuples, where the the first value 
        in the tuple is the tab title, and the second is the tab contents.
        
        title_to_set_active str: The name of the tab title to set active when the panel loads. If 
        unset, the first tab will be set active.

    Returns:
        domonic.html.div: A div html element wrapping the tabs and content.  Pass this value to 
        str() to get the string html output.
    """
    if len(content_list) == 0:
        return None

    tab_links = div(_class="kda_tab")
    root_div = div(tab_links)

    for titleContent in content_list:
        title = titleContent[0]
        content = titleContent[1]
        if content is None:
            continue
        if title_to_set_active == "":
            title_to_set_active = title
        lower_title = title.lower()

        tab_button_class = "kda_tablinks kda_active" if title_to_set_active == title else "kda_tablinks"
        tab_links.appendChild(button(title, _class=tab_button_class, _onclick=f"openTab(event, '{lower_title}')"))

        scroll_div = div(_class="kda_scroll_container").html(content)
        content_div = div(_class=f"{lower_title} kda_tabcontent").html(scroll_div)
        content_div.style.display = "block" if title_to_set_active == title else "none"
        root_div.appendChild(content_div)

    return root_div

def output_custom_css_and_javascript():
    """
    Outputs our custom css and javascript to the current cell
    """
    import IPython
    css = pkg_resources.resource_string(__name__, "formatters.css").decode("utf-8")
    js = pkg_resources.resource_string(__name__, "formatters.js").decode("utf-8")
    IPython.core.display.display(IPython.core.display.HTML(str(style(css))))
    IPython.core.display.display(IPython.core.display.HTML(str(script(js, _type="text/javascript"))))

def cell_output_is_wrapped_in_iframe():
    """
    Based on the environment that the code is running inside of, the cell output may be  
    automatically wrapped in an iFrame.  
    For example, in google colab: "the output of each cell is hosted in a separate iframe 
    sandbox with limited access to the global notebook environment"
    In enviroments like these, we need to include our custom css & javascript in every output.

    Returns:
        bool: True if the environment will automatically wrap the cell output in an iFrame, otherwise False.
    """
    if "google.colab" in str(get_ipython()):
        return True
    else:
        return False


def output_custom_css_and_javascript_if_output_wrapped_in_iframe():
    """
    Outputs our custom css and javascript to the cell ouput if the output for the cell
    will be wrapped in an iFrame.  (i.e. in google colab)
    """
    if cell_output_is_wrapped_in_iframe():
        output_custom_css_and_javascript()

def try_init():
    try:
        # the following command will throw an exception in non-iPython environments
        html_formatter = get_ipython().display_formatter.formatters["text/html"]

        # dynamically assign formatters to kaskada protobuf types
        mods = sys.modules.copy()
        for key in mods:
            if key.endswith("_grpc"):
                continue
            if key.startswith("kaskada."):
                for cls in inspect.getmembers(mods[key], inspect.isclass):
                    classname = get_classname(cls[1])
                    if classname == "kaskada.api.v1alpha.compute_pb2.QueryResponse":
                        html_formatter.for_type(classname, query_response_html_formatter) # kda_table formatter for QueryResponse
                    elif classname == "kaskada.fenl.v1alpha.schema_pb2.Schema":
                        html_formatter.for_type(classname, schema_html_formatter) 
                    elif classname == "kaskada.fenl.v1alpha.schema_pb2.DataType":
                        html_formatter.for_type(classname, data_type_html_formatter)
                    elif classname == "kaskada.prepare.v1alpha.slice_pb2.SliceRequest":
                        html_formatter.for_type(classname, slice_request_html_formatter)
                    elif classname == "kaskada.errdetails.v1alpha.fenl_diagnostics_pb2.FenlDiagnostics":
                        html_formatter.for_type(classname, fenl_diagnostics_html_formatter)
                    elif classname == "kaskada.api.v1alpha.shared_pb2.Analysis":
                        html_formatter.for_type(classname, analysis_html_formatter)
                    elif "Delete" in classname and "Response" in classname:
                        html_formatter.for_type(classname, response_delete_html_formatter) # generic formatter for Delete responses
                    elif "List" in classname and "Response" in classname:
                        html_formatter.for_type(classname, response_list_html_formatter) # generic formatter for List responses
                    elif "Response" in classname:
                        html_formatter.for_type(classname, generic_response_html_formatter) # generic formatter for other responses
                    else:
                        html_formatter.for_type(classname, generic_object_html_formatter) # generic formatter for all other objects
        
        # the following types don't normally exist when the library first loads
        html_formatter.for_type("kaskada.materialization.MaterializationView", generic_object_html_formatter)
        html_formatter.for_type("kaskada.materialization.RedisAIDestination", generic_object_html_formatter)
        html_formatter.for_type("kaskada.compute.EntityPercentFilter", entity_filter_html_formatter)
        html_formatter.for_type("fenlmagic.QueryResult", fenlmagic_query_result_html_formatter)

        # additional non-kaskada types we want to assign formatters to
        html_formatter.for_type("google.protobuf.pyext._message.RepeatedCompositeContainer", proto_list_html_formatter)

        if not cell_output_is_wrapped_in_iframe():
            # load our custom css into the page
            output_custom_css_and_javascript()

    except:
        pass



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
from typing import List
import kaskada
import kaskada.api.v1alpha.compute_pb2 as pb
from kaskada.slice_filters import EntityPercentFilter, SliceFilter

import grpc

def query(query: str, with_tables: List[pb.WithTable] = [], with_views: List[pb.WithView] = [], result_behavior: str = 'all-results', response_as: str = 'parquet', data_token_id: str = None, dry_run: bool = False, resume_token: str = None, limits: pb.QueryRequest.Limits = None, slice_filter: SliceFilter = None, client: Client = None) -> pb.QueryResponse:
    """
    Performs a query

    Args:
        query (str): The query to perform
        with_tables (List[pb.WithTable], optional): A list of tables to use in the query, in addition to the tables stored in the system.
        with_views (List[pb.WithView], optional): A list of views to use in the query, in addition to the views stored in the system.
        result_behavior (str, optional): Determines which results are returned. Either "all-results" (default), or "final-results" which returns only the final values for each entity.
        responsed_as (str, optional): Determines how the response is returned.  Either "parquet" (default) or "redis-bulk".  Note: if "redis-bulk", result_behavior is assumed to be "final-results".
        data_token_id (str, optional): Enables repeatable queries.  Queries performed against the same dataToken are always run on the same input data.
        dry_run(bool, optional): When `True`, the query is validated and if there are no errors, the resultant schema is returned. No actual computation of results is performed.
        resume_token (str, optional): Enables resumeable queries. Queries performed against a resumeToken and a dataToken produce the changed feature values from new data.
        limits (pb.QueryRequest.Limits, optional): Configure limits on the output set.
        slice_filter (SliceFilter, optional): Enables slice filter. Currently, only slice entity percent filters are supported. Defaults to None.
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        compute_pb.QueryResponse: Response from the API
    """

    if client == None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    if slice_filter is None:
        slice_filter = kaskada.KASKADA_DEFAULT_SLICE

    try:
        kaskada.validate_client(client)

        request_args = {
            "data_token_id": data_token_id,
            "dry_run": dry_run,
            "resume_token": resume_token,
            "limits": limits,
            "query" : query,
            "result_behavior": 'RESULT_BEHAVIOR_FINAL_RESULTS' if result_behavior == 'final-results' else 'RESULT_BEHAVIOR_ALL_RESULTS',
            "with_tables" : with_tables,
            "with_views" : with_views,
        }

        if response_as == 'redis-bulk':
            request_args['redis_bulk'] = {}
        else:
            request_args['parquet'] = {}

        if slice_filter is not None:
            if isinstance(slice_filter, EntityPercentFilter):
                request_args['slice'] = {
                    'percent': {
                        'percent': slice_filter.get_percent()
                    }
                }
            else:
                raise Exception('invalid slice filter provided. only EntityPercentFilter is supported')

        req = pb.QueryRequest(**request_args)

        return client.computeStub.Query(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)



class SliceFilter(object):
    pass

class EntityPercentFilter(SliceFilter):
    # The lower limit allowed for percents of entities to filter through (out of a 100)
    # The percent must be greater than or equal to this limit
    LOWER_LIMIT = 0.1
    # The upper limit allowed for percents of entities to filter through (out of a 100).
    # The percent must be less than or equal to this limit
    UPPER_LIMIT = 100.0

    def __init__(self, percent: float):
        """
        Initializes an entity percent filter. Only one filter is allowed per query request.

        Args:
            percent (float): The percent of entities to filter.
                             Must be GTE to LOWER_LIMIT and LTE to UPPER_LIMIT.
        """
        if percent < EntityPercentFilter.LOWER_LIMIT or percent > EntityPercentFilter.UPPER_LIMIT:
            raise Exception("percent must be between {} and {}".format(EntityPercentFilter.LOWER_LIMIT, EntityPercentFilter.UPPER_LIMIT))
        super().__init__()
        self.percent = percent

    def get_percent(self) -> float:
        """
        Returns the percent of entities to filter

        Returns:
            float: The percent between (0, 100)
        """
        return self.percent



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
from kaskada.slice_filters import EntityPercentFilter, SliceFilter
import kaskada.formatters

import grpc
import IPython
import os
import sys
import traceback

import kaskada.errdetails.v1alpha.fenl_diagnostics_pb2 as kaskada_error_details_pb2

from grpc_status import rpc_status
from google.rpc import error_details_pb2

# Module global client configured at initialization time. By default, all API calls utilize the global client config
KASKADA_DEFAULT_CLIENT = None
# Client ID environment variable to authenticate with the API
KASKADA_CLIENT_ID_ENV = 'KASKADA_CLIENT_ID'
# Client Secret environment variable to authenticate with the API
KASKADA_CLIENT_SECRET_ENV = 'KASKADA_CLIENT_SECRET'
# Public API endpoint for Kaskada service
KASKADA_API_ENDPOINT = 'api.kaskada.com:50051'
# Authentication endpoint for Kaskada service. JWTs exchange authentication endpoint.
KASKADA_AUTH_ENDPOINT = 'prod-kaskada.us.auth0.com'
# Authentication audience for JWTs creation.
KASKADA_AUTH_AUDIENCE = 'https://api.prod.kaskada.com'
# Demo Kaskada Client ID has read-only permissions
KASKADA_DEMO_CLIENT_ID = 'lWYFx0020u4oulh7Z9UB8C5YXVRHNyk4'
# The slicing parameter passed to query by default. By default, this is None.
KASKADA_DEFAULT_SLICE = None

def showtraceback(self, running_compiled_code=True):
    traceback_lines = traceback.format_exception(*sys.exc_info())
    message = traceback_lines[-1]
    sys.stderr.write(message)

IPython.core.interactiveshell.InteractiveShell.showtraceback = showtraceback

def init(client_id: str = None, client_secret: str = None, exchange_endpoint: str = None, audience: str = None, endpoint: str = None, is_secure: bool = True, demo: bool = False):
    """
    Initializes the Kaskada library with a client. Creating a client will update the default module default client.

    Args:
        client_id (str, optional): Kaskada Client ID.
        client_secret (str, optional): Kaskada Client Secret.
        exchange_endpoint (str, optional): Authentication endpoint to exchange JWTs.
        audience (str, optional): Authentication audience.
        endpoint (str, optional): API Endpoint.
        is_secure (bool, optional): Use SSL connection. Defaults to True.
        demo (bool, optional): Demo mode). Defaults to False.

    Returns:
        Client: Kaskada Client
    """
    kaskada.formatters.try_init()

    if demo:
        demo_only(endpoint=endpoint, is_secure=is_secure)
    else:
        client(client_id=client_id, client_secret=client_secret, exchange_endpoint=exchange_endpoint, audience=audience, endpoint=endpoint, is_secure=is_secure)



def set_default_slice(slice: SliceFilter):
    """
    Sets the default slice used in every query

    Args:
        slice (SliceFilter): SliceFilter to set the default
    """
    global KASKADA_DEFAULT_SLICE
    KASKADA_DEFAULT_SLICE = slice

def demo_only(client_id: str = None, endpoint: str = None, is_secure: bool = True):
    """
    Initializes the Kaskada library with a demo only client.

    Args:
        client_id (str, optional): The demo only client ID. Defaults to KASKADA_DEMO_CLIENT_ID.
        endpoint (str, optional): API Endpoint. Defaults to KASKADA_API_ENDPOINT.
        is_secure (bool, optional): Use SSL connection. Defaults to True.

    Returns:
        Client: Kaskada Client
    """
    if client_id is None:
        client_id = KASKADA_DEMO_CLIENT_ID

    if endpoint is None:
        endpoint = KASKADA_API_ENDPOINT

    global KASKADA_DEFAULT_CLIENT
    KASKADA_DEFAULT_CLIENT = Client.demo_only(client_id, endpoint, is_secure)
    return KASKADA_DEFAULT_CLIENT

def client(client_id: str = None, client_secret: str = None, exchange_endpoint: str = None, audience: str = None, endpoint: str = None, is_secure: bool = True) -> Client:
    """
    Initializes a Kaskada Client. Used for overloading default production parameters. Creating a client will update the default module default client.

    Args:
        client_id (str, optional): Kaskada Client ID. Defaults to environment variable KASKADA_CLIENT_ID.
        client_secret (str, optional): Kaskada Client Secret. Defaults to environment variable KASKADA_CLIENT_SECRET.
        exchange_endpoint (str, optional): Authentication endpoint to exchange JWTs. Defaults to KASKADA_AUTH_ENDPOINT.
        audience (str, optional): Authentication audience. Defaults to KASKADA_AUTH_AUDIENCE.
        endpoint (str, optional): API Endpoint. Defaults to KASKADA_API_ENDPOINT.
        is_secure (bool, optional): Use SSL connection. Defaults to True.

    Raises:
        ValueError: Provided Client ID is None.
        ValueError: Provided Client Secret is None

    Returns:
        Client: Kaskada Client
    """
    if client_id is None:
        client_id = os.getenv(KASKADA_CLIENT_ID_ENV)
    if client_id is None:
        raise ValueError("Unable to read Kaskada Client ID. Please provide client_id key word argument or set env var KASKADA_CLIENT_ID.")

    if client_secret is None:
        client_secret = os.getenv(KASKADA_CLIENT_SECRET_ENV)
    if client_secret is None:
        raise ValueError("Unable to read Kaskada Client Secret. Please provide client_secret key word argument or set env var KASKADA_CLIENT_SECRET.")

    if exchange_endpoint is None:
        exchange_endpoint = KASKADA_AUTH_ENDPOINT

    if audience is None:
        audience = KASKADA_AUTH_AUDIENCE

    if endpoint is None:
        endpoint = KASKADA_API_ENDPOINT

    global KASKADA_DEFAULT_CLIENT
    KASKADA_DEFAULT_CLIENT = Client.authorized(client_id, client_secret, exchange_endpoint, audience, endpoint, is_secure)
    return KASKADA_DEFAULT_CLIENT


def handleGrpcError(rpc_error: grpc.Call):
    """
    All methods calling `handleGrpcError` do so after detecting a RpcError.  Currently
    all our grpc calls perform a `UnaryUnaryMultiCallable` method, docs linked here:
    https://grpc.github.io/grpc/python/_modules/grpc.html#UnaryUnaryMultiCallable

    Calling a `UnaryUnaryMultiCallable` method results in one of the following:
        Returns:
            The response value for the RPC.

        Raises:
            RpcError: Indicating that the RPC terminated with non-OK status. The
            raised RpcError will also be a grpc.Call for the RPC affording the RPC's
            metadata, status code, and details.

    Methods on a `grpc.Call` object include the following.  Full docs here:
    https://grpc.github.io/grpc/python/_modules/grpc.html#Call

        code(): The StatusCode value for the RPC.
        details(): The details string of the RPC.

    The `grpc_status.rpc_status.from_call(call)` method helps extract details
    objects from `grpc.Call` objects. Docs here:
    https://grpc.github.io/grpc/python/_modules/grpc_status/rpc_status.html#from_call

        Args:
            call: A grpc.Call instance.

        Returns:
            A google.rpc.status.Status message representing the status of the RPC.

        Raises:
            ValueError: If the gRPC call's code or details are inconsistent with the
            status code and message inside of the google.rpc.status.Status.
    """
    errorMessage = "An error occurred in your request.\n\tError Code: {}\n".format(rpc_error.code().name)
    try:
        status = rpc_status.from_call(rpc_error)
        if status:
            unpacked = []
            errorMessage += "\tError Message: {}\n".format(status.message)
            for detail in status.details:
                try:
                    val = unpack_details(detail)
                    unpacked.append(val)
                except Exception:
                    # maybe report back to wren in the future
                    pass
            if len(unpacked) > 0:
                errorMessage += "Details: \n"
                for val in unpacked:
                    errorMessage += "\t{}\n".format(val)
            raise Exception(errorMessage) from None
    except ValueError:
        # maybe report back to wren in the future
        pass
    errorMessage += "\tError Message: {}\n".format(rpc_error.details())
    raise Exception(errorMessage) from None

def unpack_details(grpc_detail):
    """Unpack a grpc status detail field (which is a 'google.protobuf.Any' type).
    `Unpack()` checks the descriptor of the passed-in message object against the stored one
    and returns False if they don't match and does not attempt any unpacking; True otherwise.
        Source:
            https://github.com/protocolbuffers/protobuf/blob/master/python/google/protobuf/internal/well_known_types.py#L81
            https://github.com/protocolbuffers/protobuf/blob/master/python/google/protobuf/internal/python_message.py#L1135

        Raises:
            google.protobuf.message.DecodeError: If it can't deserialize the message object

    `Is()` checks if the `Any` message represents the given protocol buffer type.
    """
    if grpc_detail.Is(error_details_pb2.BadRequest.DESCRIPTOR):
        val = error_details_pb2.BadRequest()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.PreconditionFailure.DESCRIPTOR):
        val = error_details_pb2.PreconditionFailure()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.RetryInfo.DESCRIPTOR):
        val = error_details_pb2.RetryInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.DebugInfo.DESCRIPTOR):
        val = error_details_pb2.DebugInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.QuotaFailure.DESCRIPTOR):
        val = error_details_pb2.QuotaFailure()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.RequestInfo.DESCRIPTOR):
        val = error_details_pb2.RequestInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.ResourceInfo.DESCRIPTOR):
        val = error_details_pb2.ResourceInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.Help.DESCRIPTOR):
        val = error_details_pb2.Help()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.LocalizedMessage.DESCRIPTOR):
        val = error_details_pb2.LocalizedMessage()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(kaskada_error_details_pb2.FenlDiagnostics.DESCRIPTOR):
        val = kaskada_error_details_pb2.FenlDiagnostics()
        grpc_detail.Unpack(val)
        error_msgs = []
        for fenl_diagnostic in val.fenl_diagnostics:
            error_msgs.append(fenl_diagnostic.formatted.replace("\n", "\n\t"))
        return "\n".join(error_msgs)
    else:
        raise ValueError(grpc_detail.type_url)

def handleException(e: Exception):
    raise Exception("An exception occurred: {}".format(e)) from None

def validate_client(client: Client):
    if client == None:
        raise ValueError('Client must be provided')
    if client.computeStub == None:
        raise ValueError('Invalid client provided. Compute service stub was not initialized properly.')
    if client.viewStub == None:
        raise ValueError('Invalid client provided. View service stub was not initialized properly.')
    if client.tableStub == None:
        raise ValueError('Invalid client provided. Table service stubs was not initialized properly.')


"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express 
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from datetime import datetime, timedelta
from typing import Tuple, List

import grpc
import http.client
import json
import certifi

import kaskada.api.v1alpha.compute_pb2_grpc as compute_grpc
import kaskada.api.v1alpha.materialization_pb2_grpc as material_grpc
import kaskada.api.v1alpha.table_pb2_grpc as table_grpc
import kaskada.api.v1alpha.staged_file_pb2_grpc as staging_grpc
import kaskada.api.v1alpha.view_pb2_grpc as view_grpc

class Client(object):
    LEGACY_ENGINE = 'LEGACY'

    def getBearerToken(endpoint: str, audience: str, client_id: str, client_secret: str):
        """
        Requests a bearer token given the provided parameters

        Args:
            endpoint (str): the authentication endpoint to request a token from
            audience (str): the audience the token is requesting
            client_id (str): the user provided client id
            client_secret (str): the user provided client secret

        Returns:
            AccessToken: Kaskada Access Token
        """
        conn = http.client.HTTPSConnection(endpoint)
        payload = {
            'client_id': client_id,
            'client_secret': client_secret,
            'audience': audience, 
            'grant_type': 'client_credentials',
        }
        headers = {
            'content-type': 'application/json',
        }
        conn.request('POST', '/oauth/token', json.dumps(payload), headers)
        res = conn.getresponse()
        data = res.read().decode('utf-8')
        return Client.AccessToken(data, endpoint, audience, client_id, client_secret)

    class AccessToken(object):
        def __init__(self, json_data: str, endpoint: str, audience: str, client_id: str, client_secret: str):
            """
            AccessToken constructor

            Args:
                json_data (str): UTF-8 response from the oauth token request
                endpoint (str): the authentication endpoint to request a token from
                audience (str): the audience the token is requesting
                client_id (str): the user provided client id
                client_secret (str): the user provided client secret

            Raises:
                PermissionError: Unable to validate access token.
            """
            self.init_time = datetime.now()
            data = json.loads(json_data)
            if 'access_token' not in data and 'expires_in' not in data and 'token_type' not in data:
                raise PermissionError('Unable to validate access token. Token details: {}'.format(json_data))
            self.access_token = data['access_token']
            # The token is valid from now + expires_in seconds. Set the renewal time to one hour before the expires time for safety.
            self.expires_at = self.init_time + timedelta(0, data['expires_in'] - 3600)
            self.token_type = data['token_type']
            self.endpoint = endpoint
            self.audience = audience
            self.client_id = client_id
            self.client_secret = client_secret

    def authorized(client_id: str, client_secret: str, exchange_endpoint: str, audience: str, endpoint: str, is_secure: bool):
        """
        Initializes an authorized Kaskada Client

        Args:
            client_id (str): Kaskada Client ID
            client_secret (str): Kaskada Client Secret
            exchange_endpoint (str): Authentication endpoint to exchange JWTs
            audience (str): Authentication audience
            endpoint (str): API Endpoint
            is_secure (bool): Use SSL connection.

        Returns:
            Client: Authorized client
        """
        token = Client.getBearerToken(exchange_endpoint, audience, client_id, client_secret)
        return Client(client_id, token, endpoint, is_secure)

    def demo_only(client_id: str, endpoint: str, is_secure: bool):
        """
        Initializes a demo only Kaskada Client

        Args:
            client_id (str): Kaskada Demo only Client ID
            endpoint (str):  API Endpoint
            is_secure (bool): Use SSL
        
        Returns:
            Client: Demo client
        """
        return Client(client_id, None, endpoint, is_secure)

    def __init__(self, client_id: str, token: AccessToken, endpoint: str, is_secure: bool):
        """
        Initializes a Kaskada Client 

        Args:
            client_id (str): Kaskada Client ID
            token (AccessToken): Authentication access token provided after successful authentication
            endpoint (str): API Endpoint
            is_secure (bool): Use SSL
        """
        if is_secure:
            with open(certifi.where(), 'rb') as f:
                trusted_certs = f.read()
            credentials = grpc.ssl_channel_credentials(root_certificates=trusted_certs)
            channel = grpc.secure_channel(endpoint, credentials)
        else:
            channel = grpc.insecure_channel(endpoint)
        self.computeStub = compute_grpc.ComputeServiceStub(channel)
        self.tableStub = table_grpc.TableServiceStub(channel)
        self.viewStub = view_grpc.ViewServiceStub(channel)
        self.stagingStub = staging_grpc.StagedFileServiceStub(channel)
        self.materializationStub = material_grpc.MaterializationServiceStub(channel)
        self.client_id = client_id
        self.token = token

    def get_metadata(self) -> List[Tuple[str, str]]:
        """
        Fetches the metadata for the current client. Renews token if necessary.

        Raises:
            Exception: invalid token

        Returns:
            List[Tuple[str, str]]: Client metadata
        """
        if self.token is None:
            return [('client-id', self.client_id)]
        else:
            if self.token.expires_at is None:
                raise Exception("invalid token. please re-initialize the Kaskada client for continued use")

            current_time = datetime.now()
            remaining_token_time = (self.token.expires_at - current_time).total_seconds()
            # A token is expired if the difference between the expired time and current time is non-positive
            if (remaining_token_time <= 0):
                # Refresh the token by requesting a new one
                self.token = Client.getBearerToken(self.token.endpoint, self.token.audience, self.token.client_id, self.token.client_secret)
            return [('authorization', self.token.access_token)]
